#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <limits.h>

#include "wadstruct.h"
#include "wadio.h"
#include "lib/safeio.h"
#include "lib/safem.h"
#include "lib/log.h"
#include "lib/endian.h"

#define WDEBUG(x) logprintf(LOG_DEBUG,'W',x)

#ifndef __MSDOS__
#define O_BINARY 0
#endif

#ifdef NO_MMAP
#define DO_BIG_MMAP 0
#else
#define DO_BIG_MMAP 1
#endif

typedef struct _WadFile {
   struct _WadFile *next;
   int fd;
   int nlumps;
   const WadDirEntry *dir;
   const void *whole_map;
   const void *mapalloc;
   size_t maplen;
   char fname[PATH_MAX];
} WadFile;

#define WHOLEMAP_PTR(wad,off) ((const void *) ((const char *)(wad->whole_map)+off) )

static int nwads=0;
static WadFile *wads=NULL;

int get_num_wads(void) {return nwads;};

void reset_wad(void) {
   free_all_lumps();
   while(wads!=NULL) {
      WadFile *wad=wads;
      wads=wad->next;
      if(wad->mapalloc) safe_munmap(wad->fname,wad->mapalloc,wad->maplen);
      if(wad->fd>=0) safe_close(wad->fname,wad->fd);
      safe_free(wad);
   }
   nwads=0;
}

static void load_wad(const char *fname,const char *_sig,int mapall) {
   WadHeader whbuf;
   const WadHeader *wh=&whbuf;
   WadFile *wad=safe_calloc(sizeof(WadFile),1);
   int big_endian=*((short *)"\01\0")!=(short)1;
   int sig=*(const int *)(_sig);
   if(big_endian) sig=SWAPINT(sig);
   wad->next=wads;
   strncpy(wad->fname,fname,PATH_MAX);
   /* open the file */
   wad->fd=safe_open(fname,O_RDONLY|O_BINARY,LOG_FATAL);
   /* now get the header.  how this is done will depend on our mm strategy */
   if(mapall) {
      wad->maplen=lseek(wad->fd,0,SEEK_END);
      wh=wad->whole_map=wad->mapalloc=safe_mmap(fname,wad->fd,0,wad->maplen);
      logprintf(LOG_INFO,'W',"%s mapped at 0x%lx (%u bytes)",
		fname,(unsigned long)wh,wad->maplen); 
   }
   else
     safe_read(fname,wad->fd,&whbuf,sizeof(whbuf));
   /* deal to header */
   if(*((int *)wh->sig)==sig);
   else if(*((int *)wh->sig)==SWAPINT(sig))
     logprintf(LOG_FATAL,'W',
	       "%s is a %s-endian %s, but I'm a %s-endian processor",
	       fname,big_endian?"little":"big",_sig,big_endian?"big":"little");
   else
     logfatal('W',"Bad signature on %s: expecting %s, got %s",
	      fname,_sig,wh->sig);
   wad->nlumps=wh->nlumps;
   logprintf(LOG_INFO,'W',"%s %s: %d lumps dir @ %d",
	     _sig,fname,wad->nlumps,wh->diroffset);
   /* load directory */
   if(wad->whole_map) wad->dir=WHOLEMAP_PTR(wad,wh->diroffset);
   else wad->dir=
     safe_mmap(fname,wad->fd,wh->diroffset,wh->nlumps*sizeof(WadDirEntry));
   wads=wad;
   nwads++;
};

void init_iwad(const char *fname) {
   reset_wad();
   load_wad(fname,"IWAD",DO_BIG_MMAP);
};
void init_pwad(const char *fname) {
   load_wad(fname,"PWAD",DO_BIG_MMAP);
};

static WadFile *wadnum2file(int wadnum) {
   WadFile *w=wads;
   while(wadnum>0&&w!=NULL) {
      w=w->next;
      wadnum--;
   }
   return w;
};

int get_num_lumps(unsigned int wadnum) {
   return wadnum2file(wadnum)->nlumps;
};

static int lumpnamecmp(const char *s1,const char *s2)  {
   int i;
   for(i=0;i<8;i++)  {
      if(s2[i]!='?'&&toupper(s1[i])!=toupper(s2[i])) return 1;
      if(s2[i]==0) break;
   };
   return 0;
};

static LumpNum lumplook(LumpNum l,const char *name) {
   int wadnum=LUMP_WADNUM(l),dirnum=LUMP_DIRNUM(l);
   WadFile *wad=wadnum2file(wadnum);
   while(wad!=NULL) {
      //logprintf(LOG_DEBUG,'W',"lumplook(%s) wadnum=%d wad=0x%x",name,wadnum,wad);
      while(dirnum<wad->nlumps) {
	 if(!lumpnamecmp(wad->dir[dirnum].name,name))
	   return LUMPNUM(wadnum,dirnum);
	 dirnum++;
      };
      wadnum++;
      dirnum=0;
      wad=wad->next;
   };
   return BAD_LUMPNUM;
};

LumpNum safe_lookup_lump(const char *name,const char *after,const char *before,
			 int lvl) 
{
   LumpNum l1=0,l2=0;
   if(name==NULL||*name==0) return BAD_LUMPNUM;
   if(after) {
      l2=l1=lumplook(0,after);
      if(!LUMPNUM_OK(l1)) {
	 logprintf(lvl,'W',"Can't find lump marker %s (lump %s)",after,name);
	 return BAD_LUMPNUM;
      }
   }
   l1=lumplook(l1,name);
   if(!LUMPNUM_OK(l1)) {
      if(after) logprintf(lvl,'W',"Can't find lump %s:%s",after,name);
      else logprintf(lvl,'W',"Can't find lump %s",name);
      return BAD_LUMPNUM;
   }
   if(before) {
      l2=lumplook(l2,before);
      if(!LUMPNUM_OK(l2)) logprintf(lvl,'W',"Can't find end-marker %s",before);
      else if(l2<l1) logprintf(lvl,'W',"End-marker %s occurs before lump %s",before,name); 
   }
   return l1;
};

int count_lumps_between(const char *n1,const char *n2) {
   LumpNum ln=lumplook(0,n1),l2;
   if(!LUMPNUM_OK(ln)) return 0;
   l2=lumplook(ln,n2);
   if(!LUMPNUM_OK(l2)) return 0;
   if(LUMP_WADNUM(ln)!=LUMP_WADNUM(l2)) return 0;
   return (LUMP_DIRNUM(l2)-LUMP_DIRNUM(ln))-1;
};

static WadFile *get_lump_wad(LumpNum ln) {
   WadFile *wad=wadnum2file(LUMP_WADNUM(ln));
   if(wad==NULL) 
     logprintf(LOG_FATAL,'W',"Bad lumpnum (%x) in get_lump_*",
	       (unsigned int)ln);
   return wad;
};
static const WadDirEntry *get_lump_dirent(LumpNum ln) {
   WadFile *wad=wadnum2file(LUMP_WADNUM(ln));
   if(wad==NULL||wad->nlumps<=LUMP_DIRNUM(ln)) 
     logprintf(LOG_FATAL,'W',"Bad lumpnum (%x) in get_lump_*",
	       (unsigned int)ln);
   return wad->dir+LUMP_DIRNUM(ln);
};

int get_lump_fd(LumpNum ln) {return get_lump_wad(ln)->fd;};
unsigned int get_lump_ofs(LumpNum ln) {return get_lump_dirent(ln)->offset;};
unsigned int get_lump_len(LumpNum ln) {return get_lump_dirent(ln)->size;};
char *get_lump_name(LumpNum ln,char *d) {
   strncpy(d,get_lump_dirent(ln)->name,8);
   d[8]=0;
   return d;
};
const void *get_lump_map(LumpNum ln) {
   WadFile *wad=get_lump_wad(ln);
   const WadDirEntry *dent=get_lump_dirent(ln);
   if(wad->whole_map) return WHOLEMAP_PTR(wad,dent->offset);
   else return NULL;
}; 
